# Part of the Hex9 (H9) Project
# Copyright ©2025, Ben Griffin
# Licensed under the Apache License, Version 2.0

"""
This is the algorithms package. It is far more intuitive to import the algorithm files to be used directly.
"""
from .root_finding import find_coords

__all__ = ["find_coords"]
